#ifndef NXTDEFS__H
#define NXTDEFS__H 1
//==============================================================================
//
// Copyright (C) 2006 - John Hansen. All rights reserved.
//
// Workfile:: NXTDefs.h
// Date:: 2007-04-12
// Revision:: 24
//
//------------------------------------------------------------------------------
//
// Contains declarations for the NBC NXT API resources
//
//==============================================================================

#include "NBCCommon.h"

// define structures for various system calls

dseg	segment

TLocation	struct
 X		sword
 Y		sword
TLocation	ends

TSize	struct
 Width	sword
 Height	sword
TSize	ends

// FileOpenRead, FileOpenWrite, FileOpenAppend
TFileOpen	struct
 Result		word
 FileHandle	byte
 Filename	byte[]
 Length		dword
TFileOpen	ends

// FileRead, FileWrite
TFileReadWrite	struct
 Result		word
 FileHandle	byte
 Buffer		byte[]
 Length		dword
TFileReadWrite	ends

// FileClose
TFileClose	struct
 Result		word
 FileHandle	byte
TFileClose	ends

// FileResolveHandle
TFileResolveHandle	struct
 Result		word
 FileHandle	byte
 WriteHandle	byte
 Filename	byte[]
TFileResolveHandle	ends

// FileRename
TFileRename	struct
 Result		word
 OldFilename	byte[]
 NewFilename	byte[]
TFileRename	ends

// FileDelete
TFileDelete	struct
 Result		word
 Filename	byte[]
TFileDelete	ends

// SoundPlayFile
TSoundPlayFile	struct
 Result		sbyte
 Filename	byte[]
 Loop		byte
 Volume		byte
TSoundPlayFile	ends

// SoundPlayTone
TSoundPlayTone	struct
 Result		sbyte
 Frequency	word
 Duration	word
 Loop		byte
 Volume		byte
TSoundPlayTone	ends

// SoundGetState
TSoundGetState	struct
 State		byte
 Flags		byte
TSoundGetState	ends

// SoundSetState
TSoundSetState	struct
 Result		byte
 State		byte
 Flags		byte
TSoundSetState	ends

// DrawText
TDrawText	struct
 Result		sbyte
 Location	TLocation
 Text		byte[]
 Options	dword
TDrawText	ends

// DrawPoint
TDrawPoint	struct
 Result		sbyte
 Location	TLocation
 Options	dword
TDrawPoint	ends

// DrawLine
TDrawLine	struct
 Result		sbyte
 StartLoc	TLocation
 EndLoc		TLocation
 Options	dword
TDrawLine	ends

// DrawCircle
TDrawCircle	struct
 Result		sbyte
 Center		TLocation
 Size		byte
 Options	dword
TDrawCircle	ends

// DrawRect
TDrawRect	struct
 Result		sbyte
 Location	TLocation
 Size		TSize
 Options	dword
TDrawRect	ends

// DrawGraphic
TDrawGraphic	struct
 Result		sbyte
 Location	TLocation
 Filename	byte[]
 Variables	sdword[]
 Options	dword
TDrawGraphic	ends

// SetScreenMode
TSetScreenMode	struct
 Result		sbyte
 ScreenMode	dword
TSetScreenMode	ends

// ReadButton
TReadButton	struct
 Result		sbyte
 Index		byte
 Pressed	byte
 Count		byte
 Reset		byte
TReadButton	ends

// CommLSWrite
TCommLSWrite	struct
 Result		sbyte
 Port		byte
 Buffer		byte[]
 ReturnLen	byte
TCommLSWrite	ends

// CommLSRead
TCommLSRead	struct
 Result		sbyte
 Port		byte
 Buffer		byte[]
 BufferLen	byte
TCommLSRead	ends

// CommLSCheckStatus
TCommLSCheckStatus	struct
 Result		sbyte
 Port		byte
 BytesReady	byte
TCommLSCheckStatus	ends

// RandomNumber
TRandomNumber	struct
 Result		sword
TRandomNumber	ends

// GetStartTick
TGetStartTick	struct
 Result		dword
TGetStartTick	ends

// MessageWrite
TMessageWrite	struct
 Result		sbyte
 QueueID	byte
 Message	byte[]
TMessageWrite	ends

// MessageRead
TMessageRead	struct
 Result		sbyte
 QueueID	byte
 Remove		byte
 Message	byte[]
TMessageRead	ends

// CommBTCheckStatus
TCommBTCheckStatus	struct
 Result		sbyte
 Connection	byte
TCommBTCheckStatus	ends

// CommBTWrite
TCommBTWrite	struct
 Result		sbyte
 Connection	byte
 Buffer		byte[]
TCommBTWrite	ends

// CommBTRead
TCommBTRead	struct
 Result		sbyte
 Count		byte
 Buffer		byte[]
TCommBTRead	ends

// KeepAlive
TKeepAlive	struct
 Result		dword
TKeepAlive	ends

// IOMapRead
TIOMapRead	struct
 Result		sbyte
 ModuleName	byte[]
 Offset		word
 Count		word
 Buffer		byte[]
TIOMapRead	ends

// IOMapWrite
TIOMapWrite	struct
 Result		sbyte
 ModuleName	byte[]
 Offset		word
 Buffer		byte[]
TIOMapWrite	ends

dseg	ends

// motor arrays (compiler will optimize these out if they are not used)
dseg segment
  OUT_AB byte[] OUT_A, OUT_B
  OUT_AC byte[] OUT_A, OUT_C
  OUT_BC byte[] OUT_B, OUT_C
  OUT_ABC byte[] OUT_A, OUT_B, OUT_C
  __OnRev_Tmp sbyte
  __OnRevMutex mutex
dseg ends

#define UF_UPDATE_ONFWD 0x28

// API macros
#define ResetTachoCount(p) setout p, UpdateFlags, RESET_COUNT
#define ResetBlockTachoCount(p) setout p, UpdateFlags, RESET_BLOCK_COUNT
#define ResetRotationCount(p) setout p, UpdateFlags, RESET_ROTATION_COUNT
#define ResetAllTachoCounts(p) setout p, UpdateFlags, RESET_ALL

#define OnFwdEx(ports, pwr, reset) setout ports, Power, pwr, OutputMode, OUT_MODE_MOTORON+OUT_MODE_BRAKE, RegMode, OUT_REGMODE_IDLE, RunState, OUT_RUNSTATE_RUNNING, TachoLimit, 0, UpdateFlags, UF_UPDATE_TACHO_LIMIT+UF_UPDATE_MODE+UF_UPDATE_SPEED+reset
#define OnRevEx(ports, pwr, reset) \
  acquire __OnRevMutex \
  neg __OnRev_Tmp, pwr \
  OnFwdEx(ports, __OnRev_Tmp, reset) \
  release __OnRevMutex

#define OnFwd(ports, pwr) OnFwdEx(ports, pwr, RESET_BLOCKANDTACHO)
#define OnRev(ports, pwr) OnRevEx(ports, pwr, RESET_BLOCKANDTACHO)

#define CoastEx(ports, reset) setout ports, Power, 0, OutputMode, OUT_MODE_BRAKE, RegMode, OUT_REGMODE_IDLE, RunState, OUT_RUNSTATE_IDLE, UpdateFlags, UF_UPDATE_MODE+UF_UPDATE_SPEED+reset
#define OffEx(ports, reset) setout ports, Power, 0, OutputMode, OUT_MODE_MOTORON+OUT_MODE_BRAKE, RegMode, OUT_REGMODE_IDLE, RunState, OUT_RUNSTATE_RUNNING, UpdateFlags, UF_UPDATE_MODE+UF_UPDATE_SPEED+reset

#define Coast(ports) CoastEx(ports, RESET_BLOCKANDTACHO)
#define Off(ports) OffEx(ports, RESET_BLOCKANDTACHO)
#define Float(ports) Coast(ports)

#define OnFwdRegEx(ports, pwr, regmode, reset) setout ports, Power, pwr, OutputMode, OUT_MODE_MOTORON+OUT_MODE_REGULATED+OUT_MODE_BRAKE, RegMode, regmode, RunState, OUT_RUNSTATE_RUNNING, TachoLimit, 0, UpdateFlags, UF_UPDATE_TACHO_LIMIT+UF_UPDATE_MODE+UF_UPDATE_SPEED+reset
#define OnRevRegEx(ports, pwr, regmode, reset) \
  acquire __OnRevMutex \
  neg __OnRev_Tmp, pwr \
  OnFwdRegEx(ports, __OnRev_Tmp, regmode, reset) \
  release __OnRevMutex

#define OnFwdReg(ports, pwr, regmode) OnFwdRegEx(ports, pwr, regmode, RESET_BLOCKANDTACHO)
#define OnRevReg(ports, pwr, regmode) OnRevRegEx(ports, pwr, regmode, RESET_BLOCKANDTACHO)

#define OnFwdSyncEx(ports, pwr, turnpct, reset) setout ports, Power, pwr, OutputMode, OUT_MODE_MOTORON+OUT_MODE_REGULATED+OUT_MODE_BRAKE, RegMode, OUT_REGMODE_SYNC, TurnRatio, turnpct, RunState, OUT_RUNSTATE_RUNNING, TachoLimit, 0, UpdateFlags, UF_UPDATE_TACHO_LIMIT+UF_UPDATE_MODE+UF_UPDATE_SPEED+reset
#define OnRevSyncEx(ports, pwr, turnpct, reset) \
  acquire __OnRevMutex \
  neg __OnRev_Tmp, pwr \
  OnFwdSyncEx(ports, __OnRev_Tmp, turnpct, reset) \
  release __OnRevMutex

#define OnFwdSync(ports, pwr, turnpct) OnFwdSyncEx(ports, pwr, turnpct, RESET_BLOCKANDTACHO)
#define OnRevSync(ports, pwr, turnpct) OnRevSyncEx(ports, pwr, turnpct, RESET_BLOCKANDTACHO)

#define RotateMotor(ports, pwr, angle) \
   arrbuild __rotate_ports, ports \
   mov __rotate_power, pwr \
   mov __rotate_angle, angle \
   mov __rotate_theRVP, 40 \
   mov __rotate_theRVI, 20 \
   mov __rotate_theRVD, 100 \
   set __rotate_sync, FALSE \
   call __RotateMotor

#define RotateMotorEx(ports, pwr, angle, turnpct, bSync) \
   arrbuild __rotate_ports, ports \
   mov __rotate_power, pwr \
   mov __rotate_angle, angle \
   mov __rotate_turnpct, turnpct \
   mov __rotate_sync, bSync \
   mov __rotate_theRVP, 40 \
   mov __rotate_theRVI, 20 \
   mov __rotate_theRVD, 100 \
   call __RotateMotor

#define RotateMotorPID(ports, pwr, angle, p, i, d) \
   arrbuild __rotate_ports, ports \
   mov __rotate_power, pwr \
   mov __rotate_angle, angle \
   mov __rotate_theRVP, p \
   mov __rotate_theRVI, i \
   mov __rotate_theRVD, d \
   set __rotate_sync, FALSE \
   call __RotateMotor

#define RotateMotorExPID(ports, pwr, angle, turnpct, bSync, p, i, d) \
   arrbuild __rotate_ports, ports \
   mov __rotate_power, pwr \
   mov __rotate_angle, angle \
   mov __rotate_turnpct, turnpct \
   mov __rotate_sync, bSync \
   mov __rotate_theRVP, p \
   mov __rotate_theRVI, i \
   mov __rotate_theRVD, d \
   call __RotateMotor


subroutine __RotateMotor
// code from mot7.nbc (modified)
dseg segment
  __rotate_power byte
  __rotate_angle slong
  __rotate_ports byte[]
  __rotate_firstPort byte
  __rotate_sync byte
  __rotate_turnpct sbyte
  __rotate_theUF byte
  __rotate_theOM byte
  __rotate_theRM byte
  __rotate_theRS byte
  __rotate_theRVP byte
  __rotate_theRVI byte
  __rotate_theRVD byte
  __rotate_rs byte
  __rotate_OldRotCount sword
  __rotate_RotCount sword
  __rotate_thePower sbyte
  __rotate_theAngle ulong
  __rotate_theTurnPct sbyte
  __rotate_then dword
  __rotate_now dword
dseg ends
  // __rotate_power, __rotate_angle, and __rotate_ports have been set already
  brtst EQ, __rotate_Done, __rotate_angle
  sign __rotate_thePower, __rotate_angle
  abs __rotate_theAngle, __rotate_angle
  mul __rotate_thePower, __rotate_thePower, __rotate_power // convert __rotate_power to negative value if __rotate_angle is negative.

  set __rotate_theUF, UF_UPDATE_TACHO_LIMIT+UF_UPDATE_SPEED+UF_UPDATE_MODE+UF_UPDATE_PID_VALUES
  brtst EQ, __rotate_NoSync, __rotate_sync
  set __rotate_theOM, OUT_MODE_MOTORON+OUT_MODE_BRAKE+OUT_MODE_REGULATED
  set __rotate_theRM, OUT_REGMODE_SYNC
  mov __rotate_theTurnPct, __rotate_turnpct
  brtst EQ, __rotate_Start, __rotate_theTurnPct
  add __rotate_theUF, __rotate_theUF, UF_UPDATE_RESET_BLOCK_COUNT
  jmp __rotate_Start
__rotate_NoSync:
  set __rotate_theOM, OUT_MODE_MOTORON+OUT_MODE_BRAKE
  set __rotate_theRM, OUT_REGMODE_IDLE
  set __rotate_theTurnPct, 0
__rotate_Start:
  set __rotate_theRS, OUT_RUNSTATE_RUNNING
  setout __rotate_ports, OutputMode, __rotate_theOM, RegMode, __rotate_theRM, TachoLimit, __rotate_theAngle, RunState, __rotate_theRS, RegPValue, __rotate_theRVP, RegIValue, __rotate_theRVI, RegDValue, __rotate_theRVD, Power, __rotate_thePower, TurnRatio, __rotate_turnpct, UpdateFlags, __rotate_theUF

// Waits till angle reached
  index __rotate_firstPort, __rotate_ports, NA
__rotate_Running:
  getout __rotate_rs, __rotate_firstPort, RunState
  brcmp EQ, __rotate_Running, __rotate_rs, OUT_RUNSTATE_RUNNING
// Regulates for speed = 0
  set __rotate_theOM, OUT_MODE_MOTORON+OUT_MODE_BRAKE+OUT_MODE_REGULATED
  set __rotate_theRM, OUT_REGMODE_SPEED
  set __rotate_theUF, UF_UPDATE_SPEED+UF_UPDATE_MODE
  setout __rotate_ports, OutputMode, __rotate_theOM, RegMode, __rotate_theRM, RunState, __rotate_theRS, Power, 0, UpdateFlags, __rotate_theUF
// Verifies that motor doesn't rotate for 50ms, else loops
  getout __rotate_RotCount, __rotate_firstPort, RotationCount
__rotate_Stabilize:
  mov __rotate_OldRotCount, __rotate_RotCount
  // wait loop
  gettick __rotate_now
  add __rotate_then, __rotate_now, 50
__rotate_Waiting1:
  gettick __rotate_now
  brcmp LTEQ, __rotate_Waiting1, __rotate_now, __rotate_then 
  // check rotation
  getout __rotate_RotCount, __rotate_firstPort, RotationCount
  brcmp NEQ, __rotate_Stabilize, __rotate_OldRotCount, __rotate_RotCount
  setout __rotate_ports, RunState, OUT_RUNSTATE_IDLE, OutputMode, OUT_MODE_COAST, UpdateFlags, UF_UPDATE_MODE
  // maybe reset the block rotation count
  brtst EQ, __rotate_Done, __rotate_theTurnPct
  setout __rotate_ports, UpdateFlags, UF_UPDATE_RESET_BLOCK_COUNT
__rotate_Done:
  return
ends

dseg segment
  __SensorInvalidTmp byte
dseg ends

#define SetSensorType(port,t) setin t, port, Type
#define SetSensorMode(port,m) setin m, port, InputMode
#define ReadSensor(port,value) getin value, port, ScaledValue
#define ClearSensor(port) setin 0, port, ScaledValue

#define SetSensorTouch(port) \
  SetSensorType(port,IN_TYPE_SWITCH) \
  SetSensorMode(port,IN_MODE_BOOLEAN) \
  ResetSensor(port)

#define SetSensorLight(port) \
  SetSensorType(port,IN_TYPE_LIGHT_ACTIVE) \
  SetSensorMode(port,IN_MODE_PCTFULLSCALE) \
  ResetSensor(port)

#define SetSensorSound(port) \
  SetSensorType(port,IN_TYPE_SOUND_DB) \
  SetSensorMode(port,IN_MODE_PCTFULLSCALE) \
  ResetSensor(port)

#define SetSensorLowspeed(port) \
  SetSensorType(port,IN_TYPE_LOWSPEED_9V) \
  SetSensorMode(port,IN_MODE_RAW) \
  ResetSensor(port)

#define SetSensorUltrasonic(port) SetSensorLowspeed(port)

dseg segment
  __ResetSensorMutex mutex
  __ResetSensorPort byte
  __ResetSensorTmp byte
dseg ends

subroutine __ResetSensor
  setin TRUE, __ResetSensorPort, InvalidData
__SensorStillInvalid:
  getin	__ResetSensorTmp, __ResetSensorPort, InvalidData
  brtst	NEQ, __SensorStillInvalid, __ResetSensorTmp
  return
ends

#define ResetSensor(port) \
  acquire __ResetSensorMutex \
  mov __ResetSensorPort, port \
  call __ResetSensor \
  release __ResetSensorMutex

dseg segment
  __CLSCSArgs TCommLSCheckStatus
  __CLSCSMutex mutex
  __CLSWArgs TCommLSWrite
  __CLSWMutex mutex
  __CLSRArgs TCommLSRead
  __CLSRMutex mutex
dseg ends

#define LowspeedStatus(port, bready, result) __lowspeedStatus(port, bready, result)
#define LowspeedWrite(port, retlen, buffer, result) __lowspeedWrite(port, retlen, buffer, result)
#define LowspeedRead(port, buflen, buffer, result) __lowspeedRead(port, buflen, buffer, result)

#define __lowspeedStatus(port, bready, result) \
  acquire __CLSCSMutex \
  mov __CLSCSArgs.Port, port \
  syscall CommLSCheckStatus, __CLSCSArgs \
  mov bready, __CLSCSArgs.BytesReady \
  mov result, __CLSCSArgs.Result \
  release __CLSCSMutex

#define __lowspeedWrite(port, retlen, buffer, result) \
  acquire __CLSWMutex \
  mov __CLSWArgs.Port, port \
  mov __CLSWArgs.ReturnLen, retlen \
  mov __CLSWArgs.Buffer, buffer \
  syscall CommLSWrite, __CLSWArgs \
  mov result, __CLSWArgs.Result \
  release __CLSWMutex

#define __lowspeedRead(port, buflen, buffer, result) \
  acquire __CLSRMutex \
  mov __CLSRArgs.Port, port \
  mov __CLSRArgs.BufferLen, buflen \
  syscall CommLSRead, __CLSRArgs \
  mov buffer, __CLSRArgs.Buffer \
  mov result, __CLSRArgs.Result \
  release __CLSRMutex


dseg segment
  __TextOutMutex mutex
  __TextOutArgs TDrawText
dseg ends

#define TextOut(x,y,txt) TextOutEx(x,y,0,txt)

#define TextOutEx(x,y,cls,txt) \
  acquire __TextOutMutex \
  mov __TextOutArgs.Location.X,x \
  mov __TextOutArgs.Location.Y,y \
  mov __TextOutArgs.Options,cls \
  mov __TextOutArgs.Text,txt \
  syscall DrawText,__TextOutArgs \
  release __TextOutMutex

dseg segment
  __NumOutMutex mutex
  __NumOutArgs TDrawText
dseg ends

#define NumOut(x,y,num) NumOutEx(x,y,0,num)

#define NumOutEx(x,y,cls,num) \
  acquire __NumOutMutex \
  mov __NumOutArgs.Location.X,x \
  mov __NumOutArgs.Location.Y,y \
  mov __NumOutArgs.Options,cls \
  numtostr __NumOutArgs.Text,num \
  syscall DrawText,__NumOutArgs \
  release __NumOutMutex

dseg segment
  __PointOutArgs TDrawPoint
  __PointOutMutex mutex
dseg ends

#define PointOut(x,y) PointOutEx(x,y,0)

#define PointOutEx(x,y,cls) \
  acquire __PointOutMutex \
  mov __PointOutArgs.Location.X,x \
  mov __PointOutArgs.Location.Y,y \
  mov __PointOutArgs.Options,cls \
  syscall DrawPoint,__PointOutArgs \
  release __PointOutMutex

#define ClearScreen() PointOutEx(200, 200, 1)

dseg segment
  __LineOutArgs TDrawLine
  __LineOutMutex mutex
dseg ends

#define LineOut(x1,y1,x2,y2) LineOutEx(x1,y1,x2,y2,0)

#define LineOutEx(x1,y1,x2,y2,cls) \
  acquire __LineOutMutex \
  mov __LineOutArgs.StartLoc.X,x1 \
  mov __LineOutArgs.StartLoc.Y,y1 \
  mov __LineOutArgs.EndLoc.X,x2 \
  mov __LineOutArgs.EndLoc.Y,y2 \
  mov __LineOutArgs.Options,cls \
  syscall DrawLine,__LineOutArgs \
  release __LineOutMutex

dseg segment
  __RectOutArgs TDrawRect
  __RectOutMutex mutex
dseg ends

#define RectOut(x,y,w,h) RectOutEx(x,y,w,h,0)

#define RectOutEx(x,y,w,h,cls) \
  acquire __RectOutMutex \
  mov __RectOutArgs.Location.X,x \
  mov __RectOutArgs.Location.Y,y \
  mov __RectOutArgs.Size.Width,w \
  mov __RectOutArgs.Size.Height,h \
  mov __RectOutArgs.Options,cls \
  syscall DrawRect,__RectOutArgs \
  release __RectOutMutex

dseg segment
  __CircleOutArgs TDrawCircle
  __CircleOutMutex mutex
dseg ends

#define CircleOut(x,y,r) CircleOutEx(x,y,r,0)

#define CircleOutEx(x,y,r,cls) \
  acquire __CircleOutMutex \
  mov __CircleOutArgs.Center.X,x \
  mov __CircleOutArgs.Center.Y,y \
  mov __CircleOutArgs.Size,r \
  mov __CircleOutArgs.Options,cls \
  syscall DrawCircle,__CircleOutArgs \
  release __CircleOutMutex

dseg segment
  __GraphicOutArgs TDrawGraphic
  __GraphicOutMutex mutex
  __GraphicOutEmptyVars sdword[]
dseg ends

#define GraphicOut(x,y,file) GraphicOutEx(x,y,file,__GraphicOutEmptyVars,0)

#define GraphicOutEx(x,y,file,vars,cls) \
  acquire __GraphicOutMutex \
  mov __GraphicOutArgs.Location.X,x \
  mov __GraphicOutArgs.Location.Y,y \
  mov __GraphicOutArgs.Filename,file \
  mov __GraphicOutArgs.Variables,vars \
  mov __GraphicOutArgs.Options,cls \
  syscall DrawGraphic,__GraphicOutArgs \
  release __GraphicOutMutex

; generic I2C read routine

#define ReadI2CBytes(port, inbuf, count, outbuf, result) \
  acquire __RLSBmutex \
  mov __RLSBPort, port \
  mov __RLSReadBuf, inbuf \
  mov __RLSBytesCount, count \
  call __ReadLSBytes \
  cmp EQ, result, __RLSBytesCount, count \
  mov count, __RLSBytesCount \
  mov outbuf, __RLSReadBuf \
  release __RLSBmutex

; US Sensor read routine

#define ReadSensorUS(port,value) \
  acquire __RLSBmutex \
  mov __RLSBPort, port \
  mov __RLSReadBuf, __RLSBbufLSWrite1 \
  set __RLSBytesCount, 1 \
  call __ReadSensorLSBytesWait \
  call __ReadLSBytes \
  index value, __RLSReadBuf, NA \
  release __RLSBmutex

subroutine __ReadSensorLSBytesWait
  dseg segment
    __RLSBNow dword
    __RLSBThen dword
  dseg ends
  gettick __RLSBThen
  add __RLSBThen, __RLSBThen, 15
__RLSBWaiting:
  gettick __RLSBNow
  brcmp LT, __RLSBWaiting, __RLSBNow, __RLSBThen
  return
ends

subroutine __ReadLSBytes
  dseg segment
    __RLSBbufLSWrite1 byte[] 0x02, 0x42
    __RLSBmutex mutex
    __RLSLastGoodRead byte[] 0x00
    __RLSBPort byte
    __RLSBResult sbyte
    __RLSBytesCount byte
    __RLSBIterations byte
    __RLSReadBuf byte[]
  dseg ends
  LowspeedWrite(__RLSBPort, __RLSBytesCount, __RLSReadBuf, __RLSBResult)
  brtst EQ, __RLSBDone, __RLSBytesCount ; terminate if zero bytes to read
  arrinit __RLSReadBuf, 0, __RLSBytesCount
  brtst NEQ, __RLSBError, __RLSBResult ; terminate if not NO_ERR
  set __RLSBIterations, 4
__RLSBDoCheckStatus:
  LowspeedStatus(__RLSBPort, __RLSBytesCount, __RLSBResult)
  sub __RLSBIterations, __RLSBIterations, 1
  brtst LTEQ, __RLSBError, __RLSBIterations
  brtst LT, __RLSBError, __RLSBResult ; negative results are absolute errors
  brtst EQ, __RLSBReadyToRead, __RLSBResult
  ; if STAT_COMM_PENDING then wait a bit and then try again (up to 4 times)
  call __ReadSensorLSBytesWait
  jmp __RLSBDoCheckStatus
__RLSBReadyToRead:
  ; Try reading now
  LowspeedRead(__RLSBPort, __RLSBytesCount, __RLSReadBuf, __RLSBResult)
  brtst NEQ, __RLSBError, __RLSBResult ; terminate if not NO_ERR
  mov __RLSLastGoodRead, __RLSReadBuf
  jmp __RLSBDone
__RLSBError:
  mov __RLSReadBuf, __RLSLastGoodRead
__RLSBDone:
  arrsize __RLSBytesCount, __RLSReadBuf
  return
ends

dseg segment
  __PlayToneTmp TSoundPlayTone
  __PlayFileTmp TSoundPlayFile
dseg ends

#define PlayTone(freq,dur) PlayToneEx(freq,dur,4,0)
#define PlayFile(file) PlayFileEx(file,4,0)

#define PlayToneEx(freq,dur,vol,loop) \
  mov __PlayToneTmp.Frequency, freq \
  mov __PlayToneTmp.Duration, dur \
  mov __PlayToneTmp.Volume, vol \
  mov __PlayToneTmp.Loop, loop \
  syscall SoundPlayTone, __PlayToneTmp

#define PlayFileEx(file,vol,loop) \
  mov __PlayFileTmp.Filename, file \
  mov __PlayFileTmp.Volume, vol \
  mov __PlayFileTmp.Loop, loop \
  syscall SoundPlayFile, __PlayFileTmp

dseg segment
  __SGSMutex mutex
  __SGSArgs TSoundGetState
  __SSSMutex mutex
  __SSSArgs TSoundSetState
dseg ends

#define GetSoundState(state, flags) \
  acquire __SGSMutex \
  syscall SoundGetState, __SGSArgs \
  mov state, __SGSArgs.State \
  mov flags, __SGSArgs.Flags \
  release __SGSMutex

#define SetSoundState(state, flags, result) __setSoundState(state, flags, result)
#define __setSoundState(state, flags, result) \
  acquire __SSSMutex \
  mov __SSSArgs.State, state \
  mov __SSSArgs.Flags, flags \
  syscall SoundSetState, __SSSArgs \
  mov result, __SSSArgs.Result \
  release __SSSMutex
  
dseg segment
   __RandomTmp dword
   __RandomArgs TRandomNumber
   __RandomMutex mutex
dseg ends

#define Random(arg,max) \
   acquire __RandomMutex \
   syscall RandomNumber, __RandomArgs \
   mov __RandomTmp, __RandomArgs.Result \
   add __RandomTmp, __RandomTmp, 32768 \
   mul __RandomTmp, __RandomTmp, max \
   div __RandomTmp, __RandomTmp, 65536 \
   mov arg, __RandomTmp \
   release __RandomMutex

#define SignedRandom(arg) \
   acquire __RandomMutex \
   syscall RandomNumber, __RandomArgs \
   mov arg, __RandomArgs.Result \
   release __RandomMutex
  

dseg segment
   __KeepAliveArgs TKeepAlive
   __KeepAliveMutex mutex
dseg ends

#define ResetSleepTimer syscall KeepAlive, __KeepAliveArgs

dseg segment
  __GSTArgs TGetStartTick
  __GSTMutex mutex
dseg ends

#define GetFirstTick(value) \
  compchk EQ, sizeof(value), 4 \
  acquire __GSTMutex \
  syscall GetStartTick, __GSTArgs \
  mov value, __GSTArgs.Result \
  release __GSTMutex

dseg segment
  __RBtnMutex mutex
  __RBtnArgs TReadButton
dseg ends

#define ReadButtonEx(idx, reset, pressed, count, result) \
  acquire __RBtnMutex \
  mov __RBtnArgs.Index, idx \
  mov __RBtnArgs.Reset, reset \
  syscall ReadButton, __RBtnArgs \
  mov pressed, __RBtnArgs.Pressed \
  mov count, __RBtnArgs.Count \
  mov result, __RBtnArgs.Result \
  release __RBtnMutex

dseg segment
  __IOMRMutex mutex
  __IOMRArgs TIOMapRead
  __IOMRUnflattenErr byte
  __IOMRUnflattenBuf byte[]
dseg ends

#define GetIOMapBytes(modName, offset, cnt, arrOut) __getIOMapBytes(modName, offset, cnt, arrOut)
#define GetIOMapValue(modName, offset, n) __getIOMapValue(modName, offset, n)

#define __getIOMapBytes(modName, offset, cnt, arrOut) \
  acquire __IOMRMutex \
  mov __IOMRArgs.ModuleName, modName \
  mov __IOMRArgs.Offset, offset \
  mov __IOMRArgs.Count, cnt \
  syscall IOMapRead, __IOMRArgs \
  mov arrOut, __IOMRArgs.Buffer \
  release __IOMRMutex

#define __getIOMapValue(modName, offset, n) \
  acquire __IOMRMutex \
  mov __IOMRArgs.ModuleName, modName \
  mov __IOMRArgs.Offset, offset \
  set __IOMRArgs.Count, sizeof(n) \
  syscall IOMapRead, __IOMRArgs \
  arrtostr __IOMRUnflattenBuf, __IOMRArgs.Buffer \
  unflatten n, __IOMRUnflattenErr, __IOMRUnflattenBuf, n \
  release __IOMRMutex

#define GetCommandModuleValue(offset, n) GetIOMapValue(CommandModuleName, offset, n)
#define GetLoaderModuleValue(offset, n) GetIOMapValue(LoaderModuleName, offset, n)
#define GetSoundModuleValue(offset, n) GetIOMapValue(SoundModuleName, offset, n)
#define GetButtonModuleValue(offset, n) GetIOMapValue(ButtonModuleName, offset, n)
#define GetUIModuleValue(offset, n) GetIOMapValue(UIModuleName, offset, n)
#define GetInputModuleValue(offset, n) GetIOMapValue(InputModuleName, offset, n)
#define GetOutputModuleValue(offset, n) GetIOMapValue(OutputModuleName, offset, n)
#define GetLowSpeedModuleValue(offset, n) GetIOMapValue(LowSpeedModuleName, offset, n)
#define GetDisplayModuleValue(offset, n) GetIOMapValue(DisplayModuleName, offset, n)
#define GetCommModuleValue(offset, n) GetIOMapValue(CommModuleName, offset, n)

#define GetFreeMemory(value) \
  compchk EQ, sizeof(value), 4 \
  GetLoaderModuleValue(LoaderOffsetFreeUserFlash, value)

#define GetSoundFrequency(n) \
  compchk EQ, sizeof(n), 2 \
  GetSoundModuleValue(SoundOffsetFreq, n)
#define GetSoundDuration(n) \
  compchk EQ, sizeof(n), 2 \
  GetSoundModuleValue(SoundOffsetDuration, n)
#define GetSoundSampleRate(n) \
  compchk EQ, sizeof(n), 2 \
  GetSoundModuleValue(SoundOffsetSampleRate, n)
#define GetSoundMode(n) \
  compchk EQ, sizeof(n), 1 \
  GetSoundModuleValue(SoundOffsetMode, n)
#define GetSoundVolume(n) \
  compchk EQ, sizeof(n), 1 \
  GetSoundModuleValue(SoundOffsetVolume, n)

#define GetButtonPressCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  GetButtonModuleValue(ButtonOffsetPressedCnt(b), n)
#define GetButtonLongPressCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  GetButtonModuleValue(ButtonOffsetLongPressCnt(b), n)
#define GetButtonShortReleaseCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  GetButtonModuleValue(ButtonOffsetShortRelCnt(b), n)
#define GetButtonLongReleaseCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  GetButtonModuleValue(ButtonOffsetLongRelCnt(b), n)
#define GetButtonReleaseCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  GetButtonModuleValue(ButtonOffsetRelCnt(b), n)
#define GetButtonState(b, n) \
  compchk EQ, sizeof(n), 1 \
  GetButtonModuleValue(ButtonOffsetState(b), n)

#define GetBatteryLevel(n) \
  compchk EQ, sizeof(n), 2 \
  GetUIModuleValue(UIOffsetBatteryVoltage, n)
#define GetCommandFlags(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetFlags, n)
#define GetUIState(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetState, n)
#define GetUIButton(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetButton, n)
#define GetVMRunState(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetRunState, n)
#define GetBatteryState(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetBatteryState, n)
#define GetBluetoothState(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetBluetoothState, n)
#define GetUsbState(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetUsbState, n)
#define GetSleepTimeout(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetSleepTimeout, n)
#define GetSleepTimer(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetSleepTimer, n)
#define GetRechargeableBattery(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetSleepTimer, n)
#define GetVolume(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetVolume, n)
#define GetOnBrickProgramPointer(n) \
  compchk EQ, sizeof(n), 1 \
  GetUIModuleValue(UIOffsetOBPPointer, n)

#define GetInCustomZeroOffset(p, n) \
  compchk EQ, sizeof(n), 2 \
  GetInputModuleValue(InputOffsetCustomZeroOffset(p), n)
#define GetInSensorBoolean(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetInputModuleValue(InputOffsetSensorBoolean(p), n)
#define GetInDigiPinsDirection(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetInputModuleValue(InputOffsetDigiPinsDir(p), n)
#define GetInDigiPinsStatus(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetInputModuleValue(InputOffsetDigiPinsIn(p), n)
#define GetInDigiPinsOutputLevel(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetInputModuleValue(InputOffsetDigiPinsOut(p), n)
#define GetInCustomPercentFullScale(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetInputModuleValue(InputOffsetCustomPctFullScale(p), n)
#define GetInCustomActiveStatus(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetInputModuleValue(InputOffsetCustomActiveStatus(p), n)

#define GetOutPwnFreq(n) \
  compchk EQ, sizeof(n), 1 \
  GetOutputModuleValue(OutputOffsetPwnFreq(p), n)

#define GetLSInputBuffer(p, offset, cnt, data) GetIOMapBytes(LowSpeedModuleName, LowSpeedOffsetInBufBuf(p)+offset, cnt, data)
#define GetLSInputBufferInPtr(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetInBufInPtr(p), n)
#define GetLSInputBufferOutPtr(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetInBufOutPtr(p), n)
#define GetLSInputBufferBytesToRx(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetInBufBytesToRx(p), n)
#define GetLSOutputBuffer(p, offset, cnt, data) GetIOMapBytes(LowSpeedModuleName, LowSpeedOffsetOutBufBuf(p)+offset, cnt, data)
#define GetLSOutputBufferInPtr(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetOutBufInPtr(p), n)
#define GetLSOutputBufferOutPtr(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetOutBufOutPtr(p), n)
#define GetLSOutputBufferBytesToRx(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetOutBufBytesToRx(p), n)
#define GetLSMode(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetMode(p), n)
#define GetLSChannelState(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetChannelState(p), n)
#define GetLSErrorType(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetErrorType(p), n)
#define GetLSState(n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetState, n)
#define GetLSSpeed(n) \
  compchk EQ, sizeof(n), 1 \
  GetLowSpeedModuleValue(LowSpeedOffsetSpeed, n)

#define GetDisplayEraseMask(n) \
  compchk EQ, sizeof(n), 4 \
  GetDisplayModuleValue(DisplayOffsetEraseMask, n)
#define GetDisplayUpdateMask(n) \
  compchk EQ, sizeof(n), 4 \
  GetDisplayModuleValue(DisplayOffsetUpdateMask, n)
#define GetDisplayDisplay(n) \
  compchk EQ, sizeof(n), 4 \
  GetDisplayModuleValue(DisplayOffsetDisplay, n)
#define GetDisplayFlags(n) \
  compchk EQ, sizeof(n), 1 \
  GetDisplayModuleValue(DisplayOffsetFlags, n)
#define GetDisplayTextLinesCenterFlags(n) \
  compchk EQ, sizeof(n), 1 \
  GetDisplayModuleValue(DisplayOffsetTextLinesCenterFlags, n)
#define GetDisplayNormal(x, line, cnt, data) GetIOMapBytes(DisplayModuleName, DisplayOffsetNormal(line,x), cnt, data)
#define GetDisplayPopup(x, line, cnt, data) GetIOMapBytes(DisplayModuleName, DisplayOffsetPopup(line,x), cnt, data)

#define GetBTDeviceName(p, str) GetIOMapBytes(CommModuleName, CommOffsetBtDeviceTableName(p), 16, str)
#define GetBTDeviceClass(p, n) \
  compchk EQ, sizeof(n), 4 \
  GetCommModuleValue(CommOffsetBtDeviceTableClassOfDevice(p), n)
#define GetBTDeviceAddress(p, addr) GetIOMapBytes(CommModuleName, CommOffsetBtDeviceTableBdAddr(p), 7, addr)
#define GetBTDeviceStatus(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtDeviceTableDeviceStatus(p), n)
#define GetBTConnectionName(p, str) GetIOMapBytes(CommModuleName, CommOffsetBtConnectTableName(p), 16, str)
#define GetBTConnectionClass(p, n) \
  compchk EQ, sizeof(n), 4 \
  GetCommModuleValue(CommOffsetBtConnectTableClassOfDevice(p), n)
#define GetBTConnectionPinCode(p, code) GetIOMapBytes(CommModuleName, CommOffsetBtConnectTablePinCode(p), 16, code)
#define GetBTConnectionAddress(p, addr) GetIOMapBytes(CommModuleName, CommOffsetBtConnectTableBdAddr(p), 7, addr)
#define GetBTConnectionHandleNum(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtConnectTableHandleNr(p), n)
#define GetBTConnectionStreamStatus(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtConnectTableStreamStatus(p), n)
#define GetBTConnectionLinkQuality(p, n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtConnectTableLinkQuality(p), n)
#define GetBrickDataName(str) GetIOMapBytes(CommModuleName, CommOffsetBrickDataName, 16, str)
#define GetBrickDataBluecoreVersion(n) \
  compchk EQ, sizeof(n), 2 \
  GetCommModuleValue(CommOffsetBrickDataBluecoreVersion, n)
#define GetBrickDataAddress(p, addr) GetIOMapBytes(CommModuleName, CommOffsetBrickDataBdAddr, 7, addr)
#define GetBrickDataBtStateStatus(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBrickDataBtStateStatus, n)
#define GetBrickDataBtHardwareStatus(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBrickDataBtHwStatus, n)
#define GetBrickDataTimeoutValue(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBrickDataTimeOutValue, n)
#define GetBTInputBuffer(offset, cnt, data) GetIOMapBytes(CommModuleName, CommOffsetBtInBufBuf+offset, cnt, data)
#define GetBTInputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtInBufInPtr, n)
#define GetBTInputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtInBufOutPtr, n)
#define GetBTOutputBuffer(offset, cnt, data) GetIOMapBytes(CommModuleName, CommOffsetBtOutBufBuf+offset, cnt, data)
#define GetBTOutputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtOutBufInPtr, n)
#define GetBTOutputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtOutBufOutPtr, n)
#define GetHSInputBuffer(offset, cnt, data) GetIOMapBytes(CommModuleName, CommOffsetHsInBufBuf+offset, cnt, data)
#define GetHSInputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetHsInBufInPtr, n)
#define GetHSInputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetHsInBufOutPtr, n)
#define GetHSOutputBuffer(offset, cnt, data) GetIOMapBytes(CommModuleName, CommOffsetHsOutBufBuf+offset, cnt, data)
#define GetHSOutputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetHsOutBufInPtr, n)
#define GetHSOutputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetHsOutBufOutPtr, n)
#define GetUSBInputBuffer(offset, cnt, data) GetIOMapBytes(CommModuleName, CommOffsetUsbInBufBuf+offset, cnt, data)
#define GetUSBInputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetUsbInBufInPtr, n)
#define GetUSBInputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetUsbInBufOutPtr, n)
#define GetUSBOutputBuffer(offset, cnt, data) GetIOMapBytes(CommModuleName, CommOffsetUsbOutBufBuf+offset, cnt, data)
#define GetUSBOutputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetUsbOutBufInPtr, n)
#define GetUSBOutputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetUsbOutBufOutPtr, n)
#define GetUSBPollBuffer(offset, cnt, data) GetIOMapBytes(CommModuleName, CommOffsetUsbPollBufBuf+offset, cnt, data)
#define GetUSBPollBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetUsbPollBufInPtr, n)
#define GetUSBPollBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetUsbPollBufOutPtr, n)
#define GetBTDeviceCount(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtDeviceCnt, n)
#define GetBTDeviceNameCount(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetBtDeviceNameCnt, n)
#define GetHSFlags(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetHsFlags, n)
#define GetHSSpeed(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetHsSpeed, n)
#define GetHSState(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetHsState, n)
#define GetUSBState(n) \
  compchk EQ, sizeof(n), 1 \
  GetCommModuleValue(CommOffsetUsbState, n)

dseg segment
  __IOMWArgs TIOMapWrite
  __IOMWMutex mutex
  __IOMWFlattenBuf byte[]
dseg ends

#define SetIOMapBytes(modName, offset, cnt, arrIn) \
  acquire __IOMWMutex \
  mov __IOMWArgs.ModuleName, modName \
  mov __IOMWArgs.Offset, offset \
  arrsubset __IOMWArgs.Buffer, arrIn, NA, cnt \
  syscall IOMapWrite, __IOMWArgs \
  release __IOMWMutex

#define SetIOMapValue(modName, offset, n) \
  acquire __IOMWMutex \
  mov __IOMWArgs.ModuleName, modName \
  mov __IOMWArgs.Offset, offset \
  flatten __IOMWFlattenBuf, n \
  strtoarr __IOMWArgs.Buffer, __IOMWFlattenBuf \
  syscall IOMapWrite, __IOMWArgs \
  release __IOMWMutex

#define SetCommandModuleValue(offset, n) SetIOMapValue(CommandModuleName, offset, n)
#define SetIOCtrlModuleValue(offset, n) SetIOMapValue(IOCtrlModuleName, offset, n)
#define SetLoaderModuleValue(offset, n) SetIOMapValue(LoaderModuleName, offset, n)
#define SetUIModuleValue(offset, n) SetIOMapValue(UIModuleName, offset, n)
#define SetSoundModuleValue(offset, n) SetIOMapValue(SoundModuleName, offset, n)
#define SetButtonModuleValue(offset, n) SetIOMapValue(ButtonModuleName, offset, n)
#define SetInputModuleValue(offset, n) SetIOMapValue(InputModuleName, offset, n)
#define SetOutputModuleValue(offset, n) SetIOMapValue(OutputModuleName, offset, n)
#define SetLowSpeedModuleValue(offset, n) SetIOMapValue(LowSpeedModuleName, offset, n)
#define SetDisplayModuleValue(offset, n) SetIOMapValue(DisplayModuleName, offset, n)
#define SetCommModuleValue(offset, n) SetIOMapValue(CommModuleName, offset, n)

#define PowerDown SetIOCtrlModuleValue(IOCtrlOffsetPowerOn, IOCTRL_POWERDOWN)
#define RebootInFirmwareMode SetIOCtrlModuleValue(IOCtrlOffsetPowerOn, IOCTRL_BOOT)

#define SetSoundFrequency(n) __setSoundFrequency(n)
#define SetSoundDuration(n) __setSoundDuration(n)
#define SetSoundSampleRate(n) __setSoundSampleRate(n)
#define SetSoundFlags(n) __setSoundFlags(n)
#define SetSoundState(n) __setSoundState(n)
#define SetSoundMode(n) __setSoundMode(n)
#define SetSoundVolume(n) __setSoundVolume(n)

#define SetButtonPressCount(b, n) __setButtonPressCount(b, n)
#define SetButtonLongPressCount(b, n) __setButtonLongPressCount(b, n)
#define SetButtonShortReleaseCount(b, n) __setButtonShortReleaseCount(b, n)
#define SetButtonLongReleaseCount(b, n) __setButtonLongReleaseCount(b, n)
#define SetButtonReleaseCount(b, n) __setButtonReleaseCount(b, n)
#define SetButtonState(b, n) __setButtonState(b, n)

#define SetCommandFlags(n) __setCommandFlags(n)
#define SetUIState(n) __setUIState(n)
#define SetUIButton(n) __setUIButton(n)
#define SetVMRunState(n) __setVMRunState(n)
#define SetBatteryState(n) __setBatteryState(n)
#define SetBluetoothState(n) __setBluetoothState(n)
#define SetUsbState(n) __setUsbState(n)
#define SetSleepTimeout(n) __setSleepTimeout(n)
#define SetSleepTimer(n) __setSleepTimer(n)
#define SetVolume(n) __setVolume(n)
#define SetOnBrickProgramPointer(n) __setOnBrickProgramPointer(n)
#define ForceOff(n) __forceOff(n)

#define SetInCustomZeroOffset(p, n) __setInCustomZeroOffset(p, n)
#define SetInSensorBoolean(p, n) __setInSensorBoolean(p, n)
#define SetInDigiPinsDirection(p, n) __setInDigiPinsDirection(p, n)
#define SetInDigiPinsStatus(p, n) __setInDigiPinsStatus(p, n)
#define SetInDigiPinsOutputLevel(p, n) __setInDigiPinsOutputLevel(p, n)
#define SetInCustomPercentFullScale(p, n) __setInCustomPercentFullScale(p, n)
#define SetInCustomActiveStatus(p, n) __setInCustomActiveStatus(p, n)

#define SetOutPwnFreq(n) __setOutPwnFreq(n)

#define SetLSInputBuffer(p, offset, cnt, data) SetIOMapBytes(LowSpeedModuleName, LowSpeedOffsetInBufBuf(p)+offset, cnt, data)
#define SetLSInputBufferInPtr(p, n) __setLSInputBufferInPtr(p, n)
#define SetLSInputBufferOutPtr(p, n) __setLSInputBufferOutPtr(p, n)
#define SetLSInputBufferBytesToRx(p, n) __setLSInputBufferBytesToRx(p, n)
#define SetLSOutputBuffer(p, offset, cnt, data) SetIOMapBytes(LowSpeedModuleName, LowSpeedOffsetOutBufBuf(p)+offset, cnt, data)
#define SetLSOutputBufferInPtr(p, n) __setLSOutputBufferInPtr(p, n)
#define SetLSOutputBufferOutPtr(p, n) __setLSOutputBufferOutPtr(p, n)
#define SetLSOutputBufferBytesToRx(p, n) __setLSOutputBufferBytesToRx(p, n)
#define SetLSMode(p, n) __setLSMode(p, n)
#define SetLSChannelState(p, n) __setLSChannelState(p, n)
#define SetLSErrorType(p, n) __setLSErrorType(p, n)
#define SetLSState(n) __setLSState(n)
#define SetLSSpeed(n) __setLSSpeed(n)

#define SetDisplayEraseMask(n) __setDisplayEraseMask(n)
#define SetDisplayUpdateMask(n) __setDisplayUpdateMask(n)
#define SetDisplayDisplay(n) __setDisplayDisplay(n)
#define SetDisplayFlags(n) __setDisplayFlags(n)
#define SetDisplayTextLinesCenterFlags(n) __setDisplayTextLinesCenterFlags(n)
#define SetDisplayNormal(x, line, cnt, data) SetIOMapBytes(DisplayModuleName, DisplayOffsetNormal(line,x), cnt, data)
#define SetDisplayPopup(x, line, cnt, data) SetIOMapBytes(DisplayModuleName, DisplayOffsetPopup(line,x), cnt, data)

#define SetBTDeviceName(p, str) SetIOMapBytes(CommModuleName, CommOffsetBtDeviceTableName(p), 16, str)
#define SetBTDeviceAddress(p, addr) SetIOMapBytes(CommModuleName, CommOffsetBtDeviceTableBdAddr(p), 7, addr)
#define SetBTConnectionName(p, str) SetIOMapBytes(CommModuleName, CommOffsetBtConnectTableName(p), 16, str)
#define SetBTConnectionPinCode(p, code) SetIOMapBytes(CommModuleName, CommOffsetBtConnectTablePinCode(p), 16, code)
#define SetBTConnectionAddress(p, addr) SetIOMapBytes(CommModuleName, CommOffsetBtConnectTableBdAddr(p), 7, addr)
#define SetBrickDataName(str) SetIOMapBytes(CommModuleName, CommOffsetBrickDataName, 16, str)
#define SetBrickDataAddress(p, addr) SetIOMapBytes(CommModuleName, CommOffsetBrickDataBdAddr, 7, addr)
#define SetBTDeviceClass(p, n) __setBTDeviceClass(p, n)
#define SetBTDeviceStatus(p, n) __setBTDeviceStatus(p, n)
#define SetBTConnectionClass(p, n) __setBTConnectionClass(p, n)
#define SetBTConnectionHandleNum(p, n) __setBTConnectionHandleNum(p, n)
#define SetBTConnectionStreamStatus(p, n) __setBTConnectionStreamStatus(p, n)
#define SetBTConnectionLinkQuality(p, n) __setBTConnectionLinkQuality(p, n)
#define SetBrickDataBluecoreVersion(n) __setBrickDataBluecoreVersion(n)
#define SetBrickDataBtStateStatus(n) __setBrickDataBtStateStatus(n)
#define SetBrickDataBtHardwareStatus(n) __setBrickDataBtHardwareStatus(n)
#define SetBrickDataTimeoutValue(n) __setBrickDataTimeoutValue(n)
#define SetBTInputBuffer(offset, cnt, data) SetIOMapBytes(CommModuleName, CommOffsetBtInBufBuf+offset, cnt, data)
#define SetBTInputBufferInPtr(n) __setBTInputBufferInPtr(n)
#define SetBTInputBufferOutPtr(n) __setBTInputBufferOutPtr(n)
#define SetBTOutputBuffer(offset, cnt, data) SetIOMapBytes(CommModuleName, CommOffsetBtOutBufBuf+offset, cnt, data)
#define SetBTOutputBufferInPtr(n) __setBTOutputBufferInPtr(n)
#define SetBTOutputBufferOutPtr(n) __setBTOutputBufferOutPtr(n)
#define SetHSInputBuffer(offset, cnt, data) SetIOMapBytes(CommModuleName, CommOffsetHsInBufBuf+offset, cnt, data)
#define SetHSInputBufferInPtr(n) __setHSInputBufferInPtr(n)
#define SetHSInputBufferOutPtr(n) __setHSInputBufferOutPtr(n)
#define SetHSOutputBuffer(offset, cnt, data) SetIOMapBytes(CommModuleName, CommOffsetHsOutBufBuf+offset, cnt, data)
#define SetHSOutputBufferInPtr(n) __setHSOutputBufferInPtr(n)
#define SetHSOutputBufferOutPtr(n) __setHSOutputBufferOutPtr(n)
#define SetUSBInputBuffer(offset, cnt, data) SetIOMapBytes(CommModuleName, CommOffsetUsbInBufBuf+offset, cnt, data)
#define SetUSBInputBufferInPtr(n) __setUSBInputBufferInPtr(n)
#define SetUSBInputBufferOutPtr(n) __setUSBInputBufferOutPtr(n)
#define SetUSBOutputBuffer(offset, cnt, data) SetIOMapBytes(CommModuleName, CommOffsetUsbOutBufBuf+offset, cnt, data)
#define SetUSBOutputBufferInPtr(n) __setUSBOutputBufferInPtr(n)
#define SetUSBOutputBufferOutPtr(n) __setUSBOutputBufferOutPtr(n)
#define SetUSBPollBuffer(offset, cnt, data) SetIOMapBytes(CommModuleName, CommOffsetUsbPollBufBuf+offset, cnt, data)
#define SetUSBPollBufferInPtr(n) __setUSBPollBufferInPtr(n)
#define SetUSBPollBufferOutPtr(n) __setUSBPollBufferOutPtr(n)
#define SetBTDeviceCount(n) __setBTDeviceCount(n)
#define SetBTDeviceNameCount(n) __setBTDeviceNameCount(n)
#define SetHSFlags(n) __setHSFlags(n)
#define SetHSSpeed(n) __setHSSpeed(n)
#define SetHSState(n) __setHSState(n)
#define SetUSBState(n) __setUSBState(n)

#define __setSoundFrequency(n) \
  compchk EQ, sizeof(n), 2 \
  SetSoundModuleValue(SoundOffsetFreq, n)
#define __setSoundDuration(n) \
  compchk EQ, sizeof(n), 2 \
  SetSoundModuleValue(SoundOffsetDuration, n)
#define __setSoundSampleRate(n) \
  compchk EQ, sizeof(n), 2 \
  SetSoundModuleValue(SoundOffsetSampleRate, n)
#define __setSoundFlags(n) \
  compchk EQ, sizeof(n), 2 \
  SetSoundModuleValue(SoundOffsetFlags, n)
#define __setSoundState(n) \
  compchk EQ, sizeof(n), 1 \
  SetSoundModuleValue(SoundOffsetState, n)
#define __setSoundMode(n) \
  compchk EQ, sizeof(n), 1 \
  SetSoundModuleValue(SoundOffsetMode, n)
#define __setSoundVolume(n) \
  compchk EQ, sizeof(n), 1 \
  SetSoundModuleValue(SoundOffsetVolume, n)

#define __setButtonPressCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  SetButtonModuleValue(ButtonOffsetPressedCnt(b), n)
#define __setButtonLongPressCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  SetButtonModuleValue(ButtonOffsetLongPressCnt(b), n)
#define __setButtonShortReleaseCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  SetButtonModuleValue(ButtonOffsetShortRelCnt(b), n)
#define __setButtonLongReleaseCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  SetButtonModuleValue(ButtonOffsetLongRelCnt(b), n)
#define __setButtonReleaseCount(b, n) \
  compchk EQ, sizeof(n), 1 \
  SetButtonModuleValue(ButtonOffsetRelCnt(b), n)
#define __setButtonState(b, n) \
  compchk EQ, sizeof(n), 1 \
  SetButtonModuleValue(ButtonOffsetState(b), n)

#define __setCommandFlags(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetFlags, n)
#define __setUIState(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetState, n)
#define __setUIButton(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetButton, n)
#define __setVMRunState(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetRunState, n)
#define __setBatteryState(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetBatteryState, n)
#define __setBluetoothState(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetBluetoothState, n)
#define __setUsbState(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetUsbState, n)
#define __setSleepTimeout(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetSleepTimeout, n)
#define __setSleepTimer(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetSleepTimer, n)
#define __setVolume(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetVolume, n)
#define __setOnBrickProgramPointer(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetOBPPointer, n)
#define __forceOff(n) \
  compchk EQ, sizeof(n), 1 \
  SetUIModuleValue(UIOffsetForceOff, n)

#define __setInCustomZeroOffset(p, n) \
  compchk EQ, sizeof(n), 2 \
  SetInputModuleValue(InputOffsetCustomZeroOffset(p), n)
#define __setInSensorBoolean(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetInputModuleValue(InputOffsetSensorBoolean(p), n)
#define __setInDigiPinsDirection(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetInputModuleValue(InputOffsetDigiPinsDir(p), n)
#define __setInDigiPinsStatus(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetInputModuleValue(InputOffsetDigiPinsIn(p), n)
#define __setInDigiPinsOutputLevel(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetInputModuleValue(InputOffsetDigiPinsOut(p), n)
#define __setInCustomPercentFullScale(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetInputModuleValue(InputOffsetCustomPctFullScale(p), n)
#define __setInCustomActiveStatus(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetInputModuleValue(InputOffsetCustomActiveStatus(p), n)

#define __setOutPwnFreq(n) \
  compchk EQ, sizeof(n), 1 \
  SetOutputModuleValue(OutputOffsetPwnFreq(p), n)

#define __setLSInputBufferInPtr(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetInBufInPtr(p), n)
#define __setLSInputBufferOutPtr(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetInBufOutPtr(p), n)
#define __setLSInputBufferBytesToRx(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetInBufBytesToRx(p), n)
#define __setLSOutputBufferInPtr(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetOutBufInPtr(p), n)
#define __setLSOutputBufferOutPtr(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetOutBufOutPtr(p), n)
#define __setLSOutputBufferBytesToRx(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetOutBufBytesToRx(p), n)
#define __setLSMode(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetMode(p), n)
#define __setLSChannelState(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetChannelState(p), n)
#define __setLSErrorType(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetErrorType(p), n)
#define __setLSState(n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetState, n)
#define __setLSSpeed(n) \
  compchk EQ, sizeof(n), 1 \
  SetLowSpeedModuleValue(LowSpeedOffsetSpeed, n)

#define __setDisplayEraseMask(n) \
  compchk EQ, sizeof(n), 4 \
  SetDisplayModuleValue(DisplayOffsetEraseMask, n)
#define __setDisplayUpdateMask(n) \
  compchk EQ, sizeof(n), 4 \
  SetDisplayModuleValue(DisplayOffsetUpdateMask, n)
#define __setDisplayDisplay(n) \
  compchk EQ, sizeof(n), 4 \
  SetDisplayModuleValue(DisplayOffsetDisplay, n)
#define __setDisplayFlags(n) \
  compchk EQ, sizeof(n), 1 \
  SetDisplayModuleValue(DisplayOffsetFlags, n)
#define __setDisplayTextLinesCenterFlags(n) \
  compchk EQ, sizeof(n), 1 \
  SetDisplayModuleValue(DisplayOffsetTextLinesCenterFlags, n)

#define __setBTDeviceClass(p, n) \
  compchk EQ, sizeof(n), 4 \
  SetCommModuleValue(CommOffsetBtDeviceTableClassOfDevice(p), n)
#define __setBTDeviceStatus(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtDeviceTableDeviceStatus(p), n)
#define __setBTConnectionClass(p, n) \
  compchk EQ, sizeof(n), 4 \
  SetCommModuleValue(CommOffsetBtConnectTableClassOfDevice(p), n)
#define __setBTConnectionHandleNum(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtConnectTableHandleNr(p), n)
#define __setBTConnectionStreamStatus(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtConnectTableStreamStatus(p), n)
#define __setBTConnectionLinkQuality(p, n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtConnectTableLinkQuality(p), n)
#define __setBrickDataBluecoreVersion(n) \
  compchk EQ, sizeof(n), 2 \
  SetCommModuleValue(CommOffsetBrickDataBluecoreVersion, n)
#define __setBrickDataBtStateStatus(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBrickDataBtStateStatus, n)
#define __setBrickDataBtHardwareStatus(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBrickDataBtHwStatus, n)
#define __setBrickDataTimeoutValue(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBrickDataTimeOutValue, n)
#define __setBTInputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtInBufInPtr, n)
#define __setBTInputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtInBufOutPtr, n)
#define __setBTOutputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtOutBufInPtr, n)
#define __setBTOutputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtOutBufOutPtr, n)
#define __setHSInputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetHsInBufInPtr, n)
#define __setHSInputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetHsInBufOutPtr, n)
#define __setHSOutputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetHsOutBufInPtr, n)
#define __setHSOutputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetHsOutBufOutPtr, n)
#define __setUSBInputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetUsbInBufInPtr, n)
#define __setUSBInputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetUsbInBufOutPtr, n)
#define __setUSBOutputBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetUsbOutBufInPtr, n)
#define __setUSBOutputBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetUsbOutBufOutPtr, n)
#define __setUSBPollBufferInPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetUsbPollBufInPtr, n)
#define __setUSBPollBufferOutPtr(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetUsbPollBufOutPtr, n)
#define __setBTDeviceCount(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtDeviceCnt, n)
#define __setBTDeviceNameCount(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetBtDeviceNameCnt, n)
#define __setHSFlags(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetHsFlags, n)
#define __setHSSpeed(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetHsSpeed, n)
#define __setHSState(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetHsState, n)
#define __setUSBState(n) \
  compchk EQ, sizeof(n), 1 \
  SetCommModuleValue(CommOffsetUsbState, n)

dseg segment
  __FOMutex mutex
  __FOArgs TFileOpen
  __FCMutex mutex
  __FCArgs TFileClose
  __FRHMutex mutex
  __FRHArgs TFileResolveHandle
  __FRMutex mutex
  __FRArgs TFileRename
  __FDMutex mutex
  __FDArgs TFileDelete
dseg ends

#define CreateFile(fname, fsize, handle, result) __createFile(fname, fsize, handle, result)
#define OpenFileAppend(fname, fsize, handle, result) __openFileAppend(fname, fsize, handle, result)
#define OpenFileRead(fname, fsize, handle, result) __openFileRead(fname, fsize, handle, result)
#define CloseFile(handle, result) __closeFile(handle, result)
#define ResolveHandle(fname, handle, writeable, result) __resolveHandle(fname, handle, writeable, result)
#define RenameFile(oldname, newname, result) __renameFile(oldname, newname, result)
#define DeleteFile(fname, result) __deleteFile(fname, result)

#define Read(handle, n, result) __readValue(handle, n, result)
#define ReadLn(handle, n, result) __readLnValue(handle, n, result)
#define ReadBytes(handle, len, buf, result) __readBytes(handle, len, buf, result)
#define ReadLnString(handle, output, result) __readLnString(handle, output, result)

#define Write(handle, n, result) __writeValue(handle, n, result)
#define WriteLn(handle, n, result) __writeLnValue(handle, n, result)
#define WriteString(handle, str, cnt, result) __writeString(handle, str, cnt, result)
#define WriteLnString(handle, str, cnt, result) __writeLnString(handle, str, cnt, result)
#define WriteBytes(handle, buf, cnt, result) __writeBytes(handle, buf, cnt, result)
#define WriteBytesEx(handle, len, buf, result) __writeBytesEx(handle, len, buf, result)

#define __createFile(fname, fsize, handle, result) \
  acquire __FOMutex \
  mov __FOArgs.Filename, fname \
  mov __FOArgs.Length, fsize \
  syscall FileOpenWrite, __FOArgs \
  mov handle, __FOArgs.FileHandle \
  mov result, __FOArgs.Result \
  release __FOMutex

#define __openFileAppend(fname, fsize, handle, result) \
  acquire __FOMutex \
  mov __FOArgs.Filename, fname \
  syscall FileOpenAppend, __FOArgs \
  mov fsize, __FOArgs.Length \
  mov handle, __FOArgs.FileHandle \
  mov result, __FOArgs.Result \
  release __FOMutex

#define __openFileRead(fname, fsize, handle, result) \
  acquire __FOMutex \
  mov __FOArgs.Filename, fname \
  syscall FileOpenRead, __FOArgs \
  mov fsize, __FOArgs.Length \
  mov handle, __FOArgs.FileHandle \
  mov result, __FOArgs.Result \
  release __FOMutex
  
#define __closeFile(handle, result) \
  acquire __FCMutex \
  mov __FCArgs.FileHandle, handle \
  syscall FileClose, __FCArgs \
  mov result, __FCArgs.Result \
  release __FCMutex

#define __resolveHandle(fname, handle, writeable, result) \
  acquire __FRHMutex \
  mov __FRHArgs.Filename, fname \
  syscall FileResolveHandle, __FRHArgs \
  mov handle, __FRHArgs.FileHandle \
  mov writeable, __FRHArgs.WriteHandle \
  mov result, __FRHArgs.Result \
  release __FRHMutex

#define __renameFile(oldname, newname, result) \
  acquire __FRMutex \
  mov __FRArgs.OldFilename, oldname \
  mov __FRArgs.NewFilename, newname \
  syscall FileRename, __FRArgs \
  mov result, __FRArgs.Result \
  release __FRMutex

#define __deleteFile(fname, result) \
  acquire __FDMutex \
  mov __FDArgs.Filename, fname \
  syscall FileDelete, __FDArgs \
  mov result, __FDArgs.Result \
  release __FDMutex

dseg segment
  __FReadArgs TFileReadWrite
  __FReadTmpByte byte
  __FReadMutex mutex
  __RLSBuffer byte[]
  __RLSOutput byte[]
  __RLSReturn word
  __RLSReturnAddress byte
dseg ends

#define __readBytes(handle, len, buf, result) \
  acquire __FReadMutex \
  mov __FReadArgs.FileHandle, handle \
  mov __FReadArgs.Length, len \
  syscall FileRead, __FReadArgs \
  mov result, __FReadArgs.Result \
  strtoarr buf, __FReadArgs.Buffer \
  mov len, __FReadArgs.Length \
  release __FReadMutex

#define __readValue(handle, n, result) \
  acquire __FReadMutex \
  mov __FReadArgs.FileHandle, handle \
  set __FReadArgs.Length, sizeof(n) \
  syscall FileRead, __FReadArgs \
  mov result, __FReadArgs.Result \
  unflatten n, __FReadTmpByte, __FReadArgs.Buffer, n \
  release __FReadMutex

#define __readLnValue(handle, n, result) \
  acquire __FReadMutex \
  mov __FReadArgs.FileHandle, handle \
  set __FReadArgs.Length, sizeof(n) \
  syscall FileRead, __FReadArgs \
  unflatten n, __FReadTmpByte, __FReadArgs.Buffer, n \
  set __FReadArgs.Length, 2 \
  syscall FileRead, __FReadArgs \
  mov result, __FReadArgs.Result \
  release __FReadMutex
  
#define __readLnString(handle, output, result) \
  acquire __FReadMutex \
  mov __FReadArgs.FileHandle, handle \
  subcall __readStringLine, __RLSReturnAddress \
  mov result, __RLSReturn \
  mov output, __RLSOutput \
  release __FReadMutex \

subroutine __readStringLine
  arrinit __RLSOutput, 0, 1
  __RLSStringLoop:
  set __FReadArgs.Length, 1
  mov __RLSBuffer, __RLSOutput
  syscall FileRead, __FReadArgs
  mov __RLSReturn, __FReadArgs.Result
  brtst NEQ, __RLSStringDone, __RLSReturn
  index __FReadTmpByte, __FReadArgs.Buffer, NA
  brcmp EQ, __RLSStringDone, __FReadTmpByte, 0x0A
  brcmp EQ, __RLSStringSkip, __FReadTmpByte, 0x0D
  strcat __RLSOutput, __RLSBuffer, __FReadArgs.Buffer
  __RLSStringSkip:
  jmp __RLSStringLoop
  __RLSStringDone:
  subret __RLSReturnAddress
ends

dseg segment
  __FWriteArgs TFileReadWrite
  __FWriteFlattenBuf byte[]
  __FWriteMutex mutex
  __FWriteLn byte[] {0x0D, 0x0A}
dseg ends

#define __writeBytes(handle, buf, len, result) \
  acquire __FWriteMutex \
  mov __FWriteArgs.FileHandle, handle \
  mov __FWriteArgs.Buffer, buf \
  arrsize __FWriteArgs.Length, __FWriteArgs.Buffer \
  syscall FileWrite, __FWriteArgs \
  mov result, __FWriteArgs.Result \
  mov len, __FWriteArgs.Length \
  release __FWriteMutex

#define __writeString(handle, str, len, result) \
  acquire __FWriteMutex \
  mov __FWriteArgs.FileHandle, handle \
  strtoarr __FWriteArgs.Buffer, str \
  arrsize __FWriteArgs.Length, __FWriteArgs.Buffer \
  syscall FileWrite, __FWriteArgs \
  mov result, __FWriteArgs.Result \
  mov len, __FWriteArgs.Length \
  release __FWriteMutex

#define __writeLnString(handle, str, len, result) \
  acquire __FWriteMutex \
  mov __FWriteArgs.FileHandle, handle \
  strtoarr __FWriteFlattenBuf, str \
  arrbuild __FWriteArgs.Buffer, __FWriteFlattenBuf, __FWriteLn \
  arrsize __FWriteArgs.Length, __FWriteArgs.Buffer \
  syscall FileWrite, __FWriteArgs \
  mov result, __FWriteArgs.Result \
  mov len, __FWriteArgs.Length \
  release __FWriteMutex

#define __writeBytesEx(handle, len, buf, result) \
  acquire __FWriteMutex \
  mov __FWriteArgs.FileHandle, handle \
  mov __FWriteArgs.Length, len \
  mov __FWriteArgs.Buffer, buf \
  syscall FileWrite, __FWriteArgs \
  mov result, __FWriteArgs.Result \
  mov len, __FWriteArgs.Length \
  release __FWriteMutex

#define __writeValue(handle, n, result) \
  acquire __FWriteMutex \
  mov __FWriteArgs.FileHandle, handle \
  flatten __FWriteFlattenBuf, n \
  strtoarr __FWriteArgs.Buffer, __FWriteFlattenBuf \
  arrsize __FWriteArgs.Length, __FWriteArgs.Buffer \
  syscall FileWrite, __FWriteArgs \
  mov result, __FWriteArgs.Result \
  release __FWriteMutex

#define __writeLnValue(handle, n, result) \
  acquire __FWriteMutex \
  mov __FWriteArgs.FileHandle, handle \
  flatten __FWriteFlattenBuf, n \
  strtoarr __FWriteArgs.Buffer, __FWriteFlattenBuf \
  arrbuild __FWriteFlattenBuf, __FWriteArgs.Buffer, __FWriteLn \
  mov __FWriteArgs.Buffer, __FWriteFlattenBuf \
  arrsize __FWriteArgs.Length, __FWriteArgs.Buffer \
  syscall FileWrite, __FWriteArgs \
  mov result, __FWriteArgs.Result \
  release __FWriteMutex


dseg segment
  __MWMutex mutex
  __MWArgs TMessageWrite
  __MRMutex mutex
  __MRArgs TMessageRead
  __SRNTmpVal sdword
  __RRNTmpVal sdword
  __RRNErr byte
dseg ends

#define SendMessage(queue, msg, result) __sendMessage(queue, msg, result)
#define ReceiveMessage(queue, clear, msg, result) __receiveMessage(queue, clear, msg, result)

#define __sendMessage(queue, msg, result) \
  acquire __MWMutex \
  mov __MWArgs.QueueID, queue \
  mov __MWArgs.Message, msg \
  syscall MessageWrite, __MWArgs \
  mov result, __MWArgs.Result \
  release __MWMutex

#define __receiveMessage(queue, clear, msg, result) \
  acquire __MRMutex \
  mov __MRArgs.QueueID, queue \
  mov __MRArgs.Remove, clear \
  syscall MessageRead, __MRArgs \
  mov msg, __MRArgs.Message \
  mov result, __MRArgs.Result \
  release __MRMutex

#define ReceiveRemoteBool(queue, clear, bval, result) __receiveRemoteBool(queue, clear, bval, result)
#define ReceiveRemoteNumber(queue, clear, val, result) __receiveRemoteNumber(queue, clear, val, result)
#define ReceiveRemoteString(queue, clear, str, result) __receiveMessage(queue, clear, str, result)
#define ReceiveRemoteMessageEx(queue, clear, str, val, bval, result) __receiveRemoteMessageEx(queue, clear, str, val, bval, result)

#define __receiveRemoteBool(queue, clear, bval, result) \
  acquire __MRMutex \
  mov __MRArgs.QueueID, queue \
  mov __MRArgs.Remove, clear \
  set bval, 0 \
  syscall MessageRead, __MRArgs \
  brtst NEQ, __RRB_Err__I__, __MRArgs.Result \
  index bval, __MRArgs.Message, NA \
  __RRB_Err__I__: \
  __IncI__ \
  mov result, __MRArgs.Result \
  release __MRMutex

#define __receiveRemoteNumber(queue, clear, val, result) \
  acquire __MRMutex \
  mov __MRArgs.QueueID, queue \
  mov __MRArgs.Remove, clear \
  set __RRNTmpVal, 0 \
  syscall MessageRead, __MRArgs \
  unflatten __RRNTmpVal, __RRNErr, __MRArgs.Message, __RRNTmpVal \
  mov val, __RRNTmpVal \
  mov result, __MRArgs.Result \
  release __MRMutex

#define __receiveRemoteMessageEx(queue, clear, str, val, bval, result) \
  acquire __MRMutex \
  mov __MRArgs.QueueID, queue \
  mov __MRArgs.Remove, clear \
  set __RRNTmpVal, 0 \
  set bval, 0 \
  syscall MessageRead, __MRArgs \
  brtst NEQ, __RRM_Err__I__, __MRArgs.Result \
  index bval, __MRArgs.Message, NA \
  unflatten __RRNTmpVal, __RRNErr, __MRArgs.Message, __RRNTmpVal \
  __RRM_Err__I__: \
  __IncI__ \
  mov val, __RRNTmpVal \
  mov str, __MRArgs.Message \
  mov result, __MRArgs.Result \
  release __MRMutex

#define SendResponseString(queue, msg, result) __sendResponseString(queue, msg, result)
#define SendResponseBool(queue, bval, result) __sendResponseBool(queue, bval, result)
#define SendResponseNumber(queue, val, result) __sendResponseNumber(queue, val, result)

#define __sendResponseBool(queue, bval, result) \
  acquire __MWMutex \
  add __MWArgs.QueueID, queue, 10 \
  arrbuild __MWArgs.Message, bval, 0 \
  syscall MessageWrite, __MWArgs \
  mov result, __MWArgs.Result \
  release __MWMutex

#define __sendResponseNumber(queue, val, result) \
  acquire __MWMutex \
  add __MWArgs.QueueID, queue, 10 \
  mov __SRNTmpVal, val \
  flatten __MWArgs.Message, __SRNTmpVal \
  syscall MessageWrite, __MWArgs \
  mov result, __MWArgs.Result \
  release __MWMutex

#define __sendResponseString(queue, msg, result) \
  acquire __MWMutex \
  add __MWArgs.QueueID, queue, 10 \
  mov __MWArgs.Message, msg \
  syscall MessageWrite, __MWArgs \
  mov result, __MWArgs.Result \
  release __MWMutex


dseg segment
  __CBTCSArgs TCommBTCheckStatus
  __CBTCSMutex mutex
  __CBTWArgs TCommBTWrite
  __CBTWMutex mutex
  __SRSTmpBuf byte[]
  __SRSTmpVal sdword
  __SRSFlattenBuf byte[]
dseg ends

dseg segment
  __DCMessageWritePacket byte[] {0x80, 0x09, 0xFF, 0xFF} // append N bytes
  __DCStopProgramPacket byte[]  {0x80, 0x01}
  __DCStopSoundPacket byte[]    {0x80, 0x0c}
  __DCKeepAlivePacket byte[]    {0x80, 0x0d}
  __DCStartProgramPacket byte[] {0x80, 0x00, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
  __DCPlaySoundFilePacket byte[] {0x80, 0x02, 0xFF, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
  __DCPlayTonePacket byte[]     {0x80, 0x03, 0xFF, 0xFF, 0xFF, 0xFF}
dseg ends

#define BluetoothStatus(conn, result) __bluetoothStatus(conn, result)
#define BluetoothWrite(conn, buffer, result) __bluetoothWrite(conn, buffer, result)

#define __bluetoothStatus(conn, result) \
  acquire __CBTCSMutex \
  mov __CBTCSArgs.Connection, conn \
  syscall CommBTCheckStatus, __CBTCSArgs \
  mov result, __CBTCSArgs.Result \
  release __CBTCSMutex

#define __bluetoothWrite(conn, buffer, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  mov __CBTWArgs.Buffer, buffer \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define SendRemoteBool(conn, queue, bval, result) __sendRemoteBool(conn, queue, bval, result)
#define SendRemoteNumber(conn, queue, val, result) __sendRemoteNumber(conn, queue, val, result)
#define SendRemoteString(conn, queue, str, result) __sendRemoteString(conn, queue, str, result)

#define __sendRemoteBool(conn, queue, bval, result) \
  acquire __CBTWMutex \
  mov __SRSTmpBuf, __DCMessageWritePacket \
  replace __SRSTmpBuf, __SRSTmpBuf, 2, queue \
  replace __SRSTmpBuf, __SRSTmpBuf, 3, 2 \
  mov __CBTWArgs.Connection, conn \
  arrbuild __CBTWArgs.Buffer, __SRSTmpBuf, bval, 0 \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __sendRemoteNumber(conn, queue, val, result) \
  acquire __CBTWMutex \
  mov __SRSTmpBuf, __DCMessageWritePacket \
  replace __SRSTmpBuf, __SRSTmpBuf, 2, queue \
  replace __SRSTmpBuf, __SRSTmpBuf, 3, 5 \
  mov __SRSTmpVal, val \
  flatten __SRSFlattenBuf, __SRSTmpVal \
  mov __CBTWArgs.Connection, conn \
  arrbuild __CBTWArgs.Buffer, __SRSTmpBuf, __SRSFlattenBuf \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __sendRemoteString(conn, queue, str, result) \
  acquire __CBTWMutex \
  mov __SRSTmpBuf, __DCMessageWritePacket \
  replace __SRSTmpBuf, __SRSTmpBuf, 2, queue \
  arrsize __SRSTmpVal, str \
  replace __SRSTmpBuf, __SRSTmpBuf, 3, __SRSTmpVal \
  mov __CBTWArgs.Connection, conn \
  arrbuild __CBTWArgs.Buffer, __SRSTmpBuf, str \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define Wait(n) waitv n

#define RemoteMessageRead(conn, queue, result) __remoteMessageRead(conn, queue, result)
#define RemoteMessageWrite(conn, queue, msg, result) __sendRemoteString(conn, queue, msg, result)
#define RemoteStartProgram(conn, filename, result) __remoteStartProgram(conn, filename, result)
#define RemoteStopProgram(conn, result) __bluetoothWrite(conn, __DCStopProgramPacket, result)
#define RemotePlaySoundFile(conn, filename, bloop, result) __remotePlaySoundFile(conn, filename, bloop, result)
#define RemotePlayTone(conn, frequency, duration, result) __remotePlayTone(conn, frequency, duration, result)
#define RemoteStopSound(conn, result) __bluetoothWrite(conn, __DCStopSoundPacket, result)
#define RemoteKeepAlive(conn, result) __bluetoothWrite(conn, __DCKeepAlivePacket, result)
#define RemoteResetScaledValue(conn, port, result) __remoteResetScaledValue(conn, port, result)
#define RemoteResetMotorPosition(conn, port, brelative, result) __remoteResetMotorPosition(conn, port, brelative, result)
#define RemoteSetInputMode(conn, port, type, mode, result) __remoteSetInputMode(conn, port, type, mode, result)
#define RemoteSetOutputState(conn, port, speed, mode, regmode, turnpct, runstate, tacholimit, result) \
  __remoteSetOutputState(conn, port, speed, mode, regmode, turnpct, runstate, tacholimit, result)

#define __remoteMessageRead(conn, queue, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  add __SRSTmpVal, queue, 10 \
  arrbuild __CBTWArgs.Buffer, 0x00, 0x13, __SRSTmpVal, queue, 0x01 \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __remoteResetScaledValue(conn, port, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  arrbuild __CBTWArgs.Buffer, 0x80, 0x08, port \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __remoteResetMotorPosition(conn, port, brelative, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  arrbuild __CBTWArgs.Buffer, 0x80, 0x0a, port, brelative \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __remoteSetInputMode(conn, port, type, mode, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  arrbuild __CBTWArgs.Buffer, 0x80, 0x05, port, type, mode \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __remoteSetOutputState(conn, port, speed, mode, regmode, turnpct, runstate, tacholimit, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  mov __SRSTmpVal, tacholimit \
  flatten __SRSFlattenBuf, __SRSTmpVal \
  arrbuild __SRSTmpBuf, 0x80, 0x04, port, speed, mode, regmode, turnpct, runstate, __SRSFlattenBuf \
  strtoarr __CBTWArgs.Buffer, __SRSTmpBuf \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __remoteStartProgram(conn, filename, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  mov __CBTWArgs.Buffer, __DCStartProgramPacket \
  strsubset __SRSTmpBuf, filename, NA, 19 \
  replace __CBTWArgs.Buffer, __CBTWArgs.Buffer, 2, __SRSTmpBuf \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __remotePlaySoundFile(conn, filename, bloop, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  mov __CBTWArgs.Buffer, __DCPlaySoundFilePacket \
  strsubset __SRSTmpBuf, filename, NA, 19 \
  replace __CBTWArgs.Buffer, __CBTWArgs.Buffer, 2, bloop \
  replace __CBTWArgs.Buffer, __CBTWArgs.Buffer, 3, __SRSTmpBuf \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

#define __remotePlayTone(conn, frequency, duration, result) \
  acquire __CBTWMutex \
  mov __CBTWArgs.Connection, conn \
  mov __CBTWArgs.Buffer, __DCPlayTonePacket \
  and __SRSTmpVal, frequency, 0xff \
  replace __CBTWArgs.Buffer, __CBTWArgs.Buffer, 2, __SRSTmpVal \
  div __SRSTmpVal, frequency, 0xff \
  replace __CBTWArgs.Buffer, __CBTWArgs.Buffer, 3, __SRSTmpVal \
  and __SRSTmpVal, duration, 0xff \
  replace __CBTWArgs.Buffer, __CBTWArgs.Buffer, 4, __SRSTmpVal \
  div __SRSTmpVal, duration, 0xff \
  replace __CBTWArgs.Buffer, __CBTWArgs.Buffer, 5, __SRSTmpVal \
  syscall CommBTWrite, __CBTWArgs \
  mov result, __CBTWArgs.Result \
  release __CBTWMutex

// math functions written by Tamas Sorosy (www.sorosy.com)

// X is any integer; Y is the sqrt value (0->max); if X<0, Y is the sqrt value of absolute X
#define sqrt(X,R) __SQRT(X,R)
#define __SQRT(X,R) \
  acquire __sqrtMutex \
  mov __sqrtValue, X \
  call __sqrtSub \
  mov R, __sqrtResult \
  release __sqrtMutex

// X is any integer in degrees; Y is 100* the sin value (-100->100)
#define sin(X,R) __SIN(X,R)
#define __SIN(X,R) \
  acquire __sinMutex \
  mov __sinValue, X \
  call __sinSub \
  mov R, __sinResult \
  release __sinMutex

// X is any integer in degrees; Y is 100* the cos value (-100->100)
#define cos(X,R) __COS(X,R)
#define __COS(X,R) \
  acquire __sinMutex \
  mov __sinValue, X \
  add __sinValue, __sinValue, 90 \
  call __sinSub \
  mov R, __sinResult \
  release __sinMutex

// X is 100* the sin value (-100->100); Y is -90->90; Y is 101 if X is outside -100->100 range
#define asin(X,R) __ASIN(X,R)
#define __ASIN(X,R) \
  acquire __asinMutex \
  mov __asinValue, X \
  call __asinSub \
  mov R, __asinResult \
  release __asinMutex
                  
// X is 100* the cos value (-100->100); Y is 0->180; Y is -11 if X is outside -100->100 range
#define acos(X,R) __ACOS(X,R)
#define __ACOS(X,R) \
  acquire __asinMutex \
  mov __asinValue, X \
  call __asinSub \
  sub R, 90, __asinResult \
  release __asinMutex


// data segment
dseg segment

  // sin/cos related tables
  __sin_table sword[] 0,2,3,5,7,9,10,12,14,16,17,19,21,22,24,26,28,29,31,33,34,36,37,39,41,42,44,45,47,48,50,52,53,54,56,57,59,60,62,63,64,66,67,68,69,71,72,73,74,75,77,78,79,80,81,82,83,84,85,86,87,87,88,89,90,91,91,92,93,93,94,95,95,96,96,97,97,97,98,98,98,99,99,99,99,100,100,100,100,100,100
  __asin_table sdword[] 0,1,1,2,2,3,3,4,5,5,6,6,7,7,8,9,9,10,10,11,12,12,13,13,14,14,15,16,16,17,17,18,19,19,20,20,21,22,22,23,24,24,25,25,26,27,27,28,29,29,30,31,31,32,33,33,34,35,35,36,37,38,38,39,40,41,41,42,43,44,44,45,46,47,48,49,49,50,51,52,53,54,55,56,57,58,59,60,62,63,64,66,67,68,70,72,74,76,79,82,90

  // mutexes
  __sqrtMutex mutex
  __sinMutex mutex
  __asinMutex mutex

  // sqrt variables
  __sqrtPairs byte[]  0, 0, 0, 0, 0, 0
  __sqrtPaircount sbyte
  __sqrtValue dword
  __sqrtResult dword
  __sqrtP dword
  __sqrtR dword
  __sqrtM dword
  __sqrtN dword

  // sin variables
  __sinValue sdword
  __sinResult sdword
  __sinValueNeg byte

  // asin variables
  __asinValue sdword
  __asinResult sdword
dseg ends

subroutine __sinSub
  // move the sin to + angle
  set __sinValueNeg, FALSE
  brtst GTEQ, __sinValuePos, __sinValue

  neg __sinValue, __sinValue
  set __sinValueNeg, TRUE

__sinValuePos:
  // get the 360 mod and check which quarter the sin falls into
  mod __sinValue, __sinValue, 360
  brcmp GT, __sinQ4, __sinValue, 270
  brcmp GT, __sinQ3, __sinValue, 180
  brcmp GT, __sinQ2, __sinValue, 90

  // 1st quarter
  index __sinResult, __sin_table, __sinValue
  jmp __sinAlmostDone

__sinQ2:
  // 2nd quarter
  sub __sinValue, 180, __sinValue
  index __sinResult, __sin_table, __sinValue
  jmp __sinAlmostDone

__sinQ3:
  // 3rd quarter
  sub __sinValue, __sinValue, 180
  index __sinResult, __sin_table, __sinValue
  neg __sinResult, __sinResult
  jmp __sinAlmostDone

__sinQ4:
  // 4th quarter
  sub __sinValue, 360, __sinValue
  index __sinResult, __sin_table, __sinValue
  neg __sinResult, __sinResult
  jmp __sinAlmostDone

__sinAlmostDone:

  // if the incoming angle was <0, need to negate the result because sin(-x)=-sin(x)
  brcmp EQ, __sinDone, __sinValueNeg, FALSE
  neg __sinResult, __sinResult

__sinDone:
  return
ends


subroutine __asinSub
  // input sin value should be -1 -> 1
  brcmp GT, __asinValueBad, __asinValue, 100
  brcmp LT, __asinValueBad, __asinValue, -100

  // check if it's 0->-1
  brtst LT, __asinValueNeg, __asinValue

  // value 0->1
  index __asinResult, __asin_table, __asinValue
  jmp __asinDone

__asinValueNeg:
  // value 0->-1
  neg __asinValue, __asinValue
  index __asinResult, __asin_table, __asinValue
  neg __asinResult, __asinResult
  jmp __asinDone

__asinValueBad:
  set __asinResult, 101

__asinDone:
  return
ends

subroutine __sqrtSub
  // if the input value is 0, we're done
  set __sqrtResult, 0
  brtst EQ, __sqrtDone, __sqrtValue

  // init the paircount array
  mov __sqrtPaircount, 0
  replace __sqrtPairs, __sqrtPairs, 0, 0
  replace __sqrtPairs, __sqrtPairs, 1, 0
  replace __sqrtPairs, __sqrtPairs, 2, 0
  replace __sqrtPairs, __sqrtPairs, 3, 0
  replace __sqrtPairs, __sqrtPairs, 4, 0

__sqrtPairsLoop:
  brtst EQ, __sqrtPairsOK, __sqrtValue
  mod __sqrtN, __sqrtValue, 100
  replace __sqrtPairs, __sqrtPairs, __sqrtPaircount, __sqrtN
  div __sqrtValue, __sqrtValue, 100
  add __sqrtPaircount, __sqrtPaircount, 1

  jmp __sqrtPairsLoop

__sqrtPairsOK:
  // get the leftmost pair
  index __sqrtP, __sqrtPairs, __sqrtPaircount
  set __sqrtResult, 1

  // find the sqrt for the leftmost pair (1-9), if 0 we're not here!
__sqrtFirstLoop:
  mul __sqrtN, __sqrtResult, __sqrtResult
  brcmp GT, __sqrtFirstOK, __sqrtN, __sqrtP
  add __sqrtResult, __sqrtResult, 1
  jmp __sqrtFirstLoop

__sqrtFirstOK:
  sub __sqrtResult, __sqrtResult, 1
  // got the sqrt of the first pair in sqrtResult

  mul __sqrtN, __sqrtResult, __sqrtResult
  sub __sqrtM, __sqrtP, __sqrtN

  // in loop we get 1 new digit in sqrtResult for each pair
__sqrtBigLoop:
  sub __sqrtPaircount, __sqrtPaircount, 1

  brtst LT, __sqrtDone, __sqrtPaircount

  mul __sqrtM, __sqrtM, 100
  index __sqrtP, __sqrtPairs, __sqrtPaircount
  add __sqrtM, __sqrtM, __sqrtP

  // find the next digit
  set __sqrtN, 1

__sqrtDigitLoop:
  mul __sqrtR, __sqrtResult, 20
  add __sqrtR, __sqrtR, __sqrtN
  mul __sqrtR, __sqrtR, __sqrtN

  brcmp GT, __sqrtDigitDone, __sqrtR, __sqrtM

  add __sqrtN, __sqrtN, 1
  jmp __sqrtDigitLoop

__sqrtDigitDone:
  sub __sqrtN, __sqrtN, 1
  // got the next digit

  // calculate the new value to continue with
  mul __sqrtR, __sqrtResult, 20
  add __sqrtR, __sqrtR, __sqrtN
  mul __sqrtR, __sqrtR, __sqrtN
  sub __sqrtM, __sqrtM, __sqrtR

  // add the new digit to the end of the result in sqrtResult
  mul __sqrtResult, __sqrtResult, 10
  add __sqrtResult, __sqrtResult, __sqrtN

  jmp __sqrtBigLoop

__sqrtDone:
  return
ends

#endif // NXTDEFS__H
